import asyncio
import inspect

from .intents import Intents
from .core.http import HTTPClient, HTTPClientProtocol
from .core.gateway import GatewayClient, GatewayClientProtocol
from .core.error import DiscordError
from .core.snowflake import Snowflake

from .events.gateway_events import GatewayEvent

import logging

logger = logging.getLogger("scurrypy.client")
logger.addHandler(logging.NullHandler())

class Client:
    """Main entry point for Discord bots.
        Ties together the moving parts: gateway, HTTP and event dispatching.
    """

    token: str
    """Bot's token."""

    intents: Intents
    """Bot intents for listening to events."""

    http: HTTPClientProtocol
    """Public HTTP session for requests."""

    shards: GatewayClientProtocol
    """Shards as a list of gateways."""

    events: dict[str: list[callable]]
    """Events for the client to listen to."""

    startup_hooks: list[callable]
    """Handlers to call once before the bot starts."""

    shutdown_hooks: list[callable]
    """Handlers to call once after the bot shuts down."""

    def __init__(self,
        *,
        token: str,
        intents: Intents = Intents.DEFAULT,
        shard_count: int = 0,
        http: HTTPClientProtocol = None,
        gateway_impl: GatewayClientProtocol = None
    ):
        """
        Args:
            token (str): the bot's token
            intents (Intents, optional): gateway intents. Defaults to `Intents.DEFAULT`.
            shard_count (int, optional): number of shards to spawn. Defaults to `0` or recommended shard count.
            http (HTTPClientProtocol, optional): HTTP protocol implementation. Leave blank for default client.
            gateway_impl (GatewayClientProtocol, optional): Gateway protocol implementation. Leave blank for default client.
        """
        if not isinstance(intents, Intents):
            raise ValueError("Invalid intents type.")
        
        self.token = token
        self.intents = intents
        self.shard_count = shard_count
        
        self.http = http or HTTPClient()

        self.shards: list[GatewayClientProtocol] = []
        self.shard_type = gateway_impl or GatewayClient

        self.events = {}
        self.startup_hooks = []
        self.shutdown_hooks = []

    def add_event_listener(self, event: str, handler):
        """Helper function to register listener functions.

        Args:
            event (str): name of the event to listen
            handler (callable): listener function
        """
        if not callable(handler):
            raise TypeError(f"{handler} is not a callable function.")
        
        params_len = len(inspect.signature(handler).parameters)

        if params_len != 1:
            raise TypeError(
                f"Event listener '{handler.__name__}' must accept exactly one parameter (event)."
            )
    
        self.events.setdefault(event, []).append(handler)

    def add_startup_hook(self, handler):
        """Helper function to register startup functions.
            Runs once on startup BEFORE READY event.

        Args:
            handler (callable): startup function
        """
        if not callable(handler):
            raise TypeError(f"{handler} is not a callable function.")
        
        params_len = len(inspect.signature(handler).parameters)

        if params_len != 0:
            raise TypeError(
                f"Startup hook '{handler.__name__}' must accept no parameters."
            )
        
        self.startup_hooks.append(handler)

    def add_shutdown_hook(self, handler):
        """Helper function to register shutdown functions.
            Runs once on shutdown.

        Args:
            handler (callable): shutdown function
        """
        if not callable(handler):
            raise TypeError(f"{handler} is not a callable function.")
        
        params_len = len(inspect.signature(handler).parameters)

        if params_len != 0:
            raise TypeError(
                f"Shutdown hook '{handler.__name__}' must accept no parameters."
            )

        self.shutdown_hooks.append(handler)

    def application(self, application_id: Snowflake):
        """Creates an interactable application resource.

        Args:
            application_id (Snowflake): ID of target application

        Returns:
            (Application): the Application resource
        """
        from .resources.application import Application

        return Application(self.http, id=application_id, context=None)
    
    def bot_emoji(self, application_id: Snowflake):
        """Creates an interactable bot emoji resource.

        Args:
            application_id (Snowflake): ID of target application

        Returns:
            (BotEmojis): the BotEmoji resource
        """
        from .resources.emoji import ApplicationEmoji

        return ApplicationEmoji(self.http, None, application_id)
    
    def guild_emoji(self, guild_id: Snowflake):
        """Creates an interactable emoji resource.

        Args:
            guild_id (Snowflake): guild ID of target emojis

        Returns:
            (GuildEmoji): the GuildEmoji resource
        """
        from .resources.emoji import GuildEmoji

        return GuildEmoji(self.http, None, guild_id)

    def guild(self, guild_id: Snowflake):
        """Creates an interactable guild resource.

        Args:
            guild_id (Snowflake): ID of target guild

        Returns:
            (Guild): the Guild resource
        """
        from .resources.guild import Guild

        return Guild(self.http, guild_id)

    def channel(self, channel_id: Snowflake):
        """Creates an interactable guild channel resource.

        Args:
            channel_id (Snowflake): ID of target channel

        Returns:
            (GuildChannel): the GuildChannel resource
        """
        from .resources.channel import Channel

        return Channel(self.http, channel_id)
    
    def invite(self, code: str):
        """Creates an interactable invite resource.

        Args:
            code (str): unique invite code
        """
        from .resources.invite import Invite

        return Invite(self.http, code)
    
    def global_command(self, application_id: Snowflake):
        """Creates an interactable command resource.

        Args:
            application_id (Snowflake): bot's user ID

        Returns:
            (GlobalCommand): the GlobalCommand resource
        """
        from .resources.command import GlobalCommand

        return GlobalCommand(self.http, application_id)
    
    def guild_command(self, application_id: Snowflake, guild_id: Snowflake = None):
        """Creates an interactable command resource.

        Args:
            application_id (Snowflake): bot's user ID
            guild_id (Snowflake, optional): ID of guild if command is in guild scope

        Returns:
            (GuildCommand): the GuildCommand resource
        """
        from .resources.command import GuildCommand

        return GuildCommand(self.http, application_id, guild_id)

    def message(self, channel_id: Snowflake, message_id: Snowflake):
        """Creates an interactable message resource.

        Args:
            message_id (Snowflake): ID of target message
            channel_id (Snowflake): channel ID of target message

        Returns:
            (Message): the Message resource
        """
        from .resources.message import Message

        return Message(self.http, message_id, channel_id)
    
    def interaction(self, id: Snowflake, token: str):
        """Creates an interactable interaction resource.

        Args:
            id (Snowflake): ID of the interaction
            token (str): interaction token

        Returns:
            (Interaction): the Interaction resource
        """
        from .resources.interaction import Interaction

        return Interaction(self.http, id, token)
    
    def sticker(self):
        """Creates an interactable sticker resource

        Returns:
            (Sticker): the Sticker resource
        """
        from .resources.sticker import Sticker

        return Sticker(self.http)
    
    def user(self):
        """Creates an interactable user resource.

        Returns:
            (User): the User resource
        """
        from .resources.user import User

        return User(self.http)

    async def listen_shard(self, shard: GatewayClient):
        """Consume a GatewayClient's event queue.

        Args:
            shard (GatewayClient): gateway to listen on
        """

        while True:
            try:
                dispatch_type, event_data = await shard.event_queue.get()

                if dispatch_type not in self.events.keys():
                    logger.debug(f"SHARD ID {shard.shard_id} DISPATCH -> {dispatch_type}")
                else:
                    logger.info(f"SHARD ID {shard.shard_id} DISPATCH -> {dispatch_type}")

                from .core.events import EVENTS
                event_model = EVENTS.get(dispatch_type)
                if not event_model:
                    logger.warning(f"Event {dispatch_type} is not implemented.")
                    continue

                obj = event_model.from_dict(event_data)
                obj.name = dispatch_type
                obj.raw = event_data

                handlers = self.events.get(dispatch_type, [])
                for handler in handlers:
                    try:
                        result = handler(obj)
                        if inspect.isawaitable(result):
                            await result
                    except DiscordError as e:
                        logger.error(e)
                        continue

            except Exception:
                # catastrophic errors (network, shard death, unexpected OP code)
                logger.exception(f"SHARD ID {shard.shard_id}: Dispatcher error")
                continue

    async def start_shards(self, gateway: GatewayEvent):
        """Starts all shards batching by max_concurrency."""

        # pull important values for easier access
        total_shards = self.shard_count or gateway.shards
        batch_size = gateway.session_start_limit.max_concurrency

        tasks = []
        
        for batch_start in range(0, total_shards, batch_size):
            batch_end = min(batch_start + batch_size, total_shards)

            logger.debug(f"Starting shards {batch_start}-{batch_end} of {total_shards}")

            for shard_id in range(batch_start, batch_end):
                shard = self.shard_type()
                self.shards.append(shard)

                # fire and forget
                tasks.append(asyncio.create_task(shard.start(self.token, self.intents, shard_id, total_shards)))
                tasks.append(asyncio.create_task(self.listen_shard(shard)))

            # wait before next batch to respect identify rate limit
            await asyncio.sleep(5)

        return tasks
    
    async def run_startup_hooks(self):
        for hook in self.startup_hooks:
            try:
                logger.debug(f"Running hook {hook.__qualname__}...")
                result = hook()
                if inspect.isawaitable(result):
                    await asyncio.wait_for(result, timeout=60)
            except Exception:
                logger.exception("Error in shartup hook")
    
    async def start(self):
        """Starts the HTTP/Websocket client, run startup logic, and registers commands."""
        
        try:
            await self.http.start(self.token)

            data = await self.http.request('GET', '/gateway/bot')

            if not data:
                return

            gateway = GatewayEvent.from_dict(data)

            await self.run_startup_hooks()

            tasks = await asyncio.create_task(self.start_shards(gateway))

            await asyncio.gather(*tasks)
            
        except asyncio.CancelledError:
            logger.info("Connection cancelled via KeyboardInterrupt.")
        except DiscordError as e:
            logger.error(e)
        except Exception:
            logger.exception(f"Unhandled client start exception.")
        finally:
            await self.close()

    async def run_shutdown_hooks(self):
        for hook in self.shutdown_hooks:
            try:
                logger.debug(f"Running hook {hook.__qualname__}...")
                result = hook()
                if inspect.isawaitable(result):
                    await asyncio.wait_for(result, timeout=60)
            except Exception:
                logger.exception("Error in shutdown hook")

    async def close(self):
        """Gracefully close HTTP session, websocket connections, and run shutdown logic."""  

        await self.run_shutdown_hooks()

        # close each connection or shard BEFORE HTTP
        await asyncio.gather(
            *(shard.close_ws() for shard in self.shards),
            return_exceptions=True
        )

        logger.info("Closing HTTP session...")
        await self.http.close()
    
    def run(self):
        """User-facing entry point for starting the client."""  

        try:
            asyncio.run(self.start())
        except Exception as e:
            logger.exception(f"{type(e).__name__} {e}")
        finally:
            logger.info("Bot shutting down.")



from dataclasses import dataclass, field

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...bases.channel import GuildChannelCreate

from ...enums.channel import ChannelType, SortOrderType, ForumLayoutType

from .channel import ChannelModel
from .default_reaction import DefaultReactionModel, DefaultReactionPart
from .tag import TagModel, TagPart

@dataclass
class GuildForumChannelModel(ChannelModel):
    """Represents the forum channel."""

    available_tags: list[TagModel] | None
    """Set of tags that can be applied to a `GUILD_FORUM` post."""

    applied_tags: list[Snowflake] | None
    """Set of tags applied to a `GUILD_FORUM` post."""

    default_reaction_emoji: DefaultReactionModel | None
    """Emoji to show in the add reaction button in a `GUILD_FORUM` post."""

    default_sort_order: SortOrderType | None
    """Default forum sort order."""

    default_forum_layout: ForumLayoutType | None
    """Default forum layout view. Discord defaults to `ForumLayoutTypes.NOT_SET`."""

@dataclass
class GuildForumChannelPart(DataModel, GuildChannelCreate):
    """Parameters for creating a guild forum channel."""

    name: str | None = None
    """Name of the channel."""

    topic: str | None = None
    """Topic of the channel."""

    position: int | None = None
    """Sorting position of the channel (channels with the same position are sorted by id)."""

    rate_limit_per_user: int | None = None
    """Seconds user must wait between sending messages in the channel."""

    parent_id: Snowflake | None = None
    """Category ID of the channel."""

    nsfw: bool | None = None
    """If the channel is flagged NSFW."""

    default_auto_archive_duration: int | None = None
    """Default duration in minutes threads will be hidden after period of inactivity."""

    default_reaction_emoji: DefaultReactionPart | None = None
    """Emoji to show in the add reaction button in a `GUILD_FORUM` post."""

    available_tags: list[TagPart] | None = None
    """Set of tags that can be applied to a `GUILD_FORUM` post."""

    default_sort_order: SortOrderType | None = None
    """Default forum sort order."""

    default_forum_layout: ForumLayoutType | None = None
    """Default forum layout view."""

    default_thread_rate_limit_per_user: int | None = None
    """Rate limit per user set on newly created threads.
    
    !!! note
        This field does not live update!
    """

    type: ChannelType = field(init=False, default=ChannelType.GUILD_FORUM)
    """Type of channel. Always `ChannelType.GUILD_FORUM` for this class."""



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from ..enums.permissions import Permissions

@dataclass
class UserModel(DataModel):
    """Represents the User object."""

    id: Snowflake
    """ID of the user."""

    username: str
    """Username of the user."""

    discriminator: str
    """Discriminator of the user (#XXXX)"""

    global_name: str
    """Global name of the user."""

    avatar: str
    """Image hash of the user's avatar."""

    bot: bool | None
    """If the user is a bot."""

    banner: str | None
    """Image hash of the user's banner."""

    accent_color: int | None
    """Color of user's banner represented as an integer."""

    locale: str | None
    """Chosen language option of the user."""

@dataclass
class GuildMemberModel(DataModel):
    """Represents a guild member."""

    roles: list[Snowflake]
    """List of roles registered to the guild member."""

    user: UserModel
    """User data associated with the guild member."""

    nick: str
    """Server nickname of the guild member."""

    avatar: str
    """Server avatar hash of the guild mmeber."""

    joined_at: str
    """ISO8601 timestamp of when the guild member joined server."""

    permissions: Permissions | None
    """Total permissions of the member in the channel, including overwrites, 
        returned when in the interaction object. [`INT_LIMIT`]
    """



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ..guilds.role import GuildRoleModel
from ..channels.channel import ChannelModel
from ..messages.message import MessageModel
from ..messages.attachment import AttachmentModel

from ..user import UserModel, GuildMemberModel

@dataclass
class ResolvedDataModel(DataModel):
    """Represents the resolved data object."""

    users: dict[Snowflake, UserModel] | None
    """Map of user snowflakes to user objects."""

    members: dict[Snowflake, GuildMemberModel] | None
    """Map of member snowflakes to partial guild member objects.

    !!! note "Missing Fields"
        `user`, `deaf`, and `mute`.
    """

    roles: dict[Snowflake, GuildRoleModel] | None
    """Map of role snowflakes to role objects."""

    channels: dict[Snowflake, ChannelModel] | None
    """Map of channel snowflakes to partial channel objects."""

    messages: dict[Snowflake, MessageModel] | None
    """Map of message snowflakes to partial message objects."""

    attachments: dict[Snowflake, AttachmentModel] | None
    """Map of attachment snowflakes to attachment objects."""



import logging

logger = logging.getLogger('scurrypy')

from scurrypy import Client, Intents
from scurrypy.bases import Addon
from scurrypy.enums import EventType
from scurrypy.core import DiscordError, Snowflake
from scurrypy.events import MessageCreateEvent

from .ctx import PrefixCommandContext

class PrefixAddon(Addon):
    """Addon that implements automatic registering and decorating prefix commands."""

    def __init__(self, client: Client, application_id: Snowflake, prefix: str):
        """
        Args:
            client (Client): the Client object
            application_id (Snowflake): ID of the bot
            prefix (str): message prefix for commands
        """
        if not Intents.MESSAGE_CONTENT in client.intents:
            raise ValueError("Missing Intent.MESSAGE_CONTENT for scanning messages.")
        
        self.bot = client

        self.application_id = application_id

        self._prefix = prefix

        self._commands = {}
        """Maps prefix command names to handler."""
        
        client.add_event_listener(EventType.MESSAGE_CREATE, self.dispatch)

    def _register(self, name: str, func: callable):
        import inspect

        params_len = len(inspect.signature(func).parameters)
        if params_len != 1:
            raise TypeError(
                f"Prefix handler '{func.__name__}' must accept exactly one parameter (ctx)."
            )

        self._commands[name.lower()] = func
        logger.info(f"Prefix command '{self._prefix + name}' registered.")

    def listen(self, name: str, *, handler: callable = None):
        """Listen for a prefix command.

        Args:
            name (str): name of the command
                !!! warning "Important"
                    Prefix commands are CASE-INSENSITIVE.
            handler (callable): callback for the command (if not a decorator)

        Raises:
            (TypeError): invalid handler signature
        """
        if handler is not None:
            self._register(name, handler)
        else:
            def decorator(func):
                self._register(name, func)
                return func
            return decorator

    async def dispatch(self, event: MessageCreateEvent):
        """Dispatch event to user-defined handler.
            Ignore bot responding to self and messages without the desired prefix.

        Args:
            event (MessageCreateEvent): message create event object
        """
        if not event.content:
            return # ignore empty messages
        
        if event.author.id == self.application_id:
            return # ignore bot responding to itself
        
        has_prefix = event.content.lower().startswith(self._prefix.lower())

        if not has_prefix:
            return # ignore messages without prefix
        
        command, *args = event.content[len(self._prefix):].strip().lower().split()
        handler = self._commands.get(command)

        # warn if this command doesnt have a known handler
        if not handler:
            logger.warning(f"Prefix Event '{command}' not found.")
            return

        # now prefix info can be confidently set
        try:
            ctx = PrefixCommandContext(self.bot, event, list(args))
            await handler(ctx)
            
            logger.info(f"Prefix Event '{self._prefix + command}' acknowledged with args: {ctx.args or 'No args'}")
        except DiscordError as e:
            logger.error(f"Error in prefix command '{self._prefix + command}': {e}")        
        except Exception as e:
            logger.exception(f"Unhandled error in prefix command '{self._prefix + command}': {e}")



from dataclasses import dataclass

from ...core.snowflake import Snowflake

from ..user import UserModel

from .channel import ChannelModel

@dataclass
class DMChannelModel(ChannelModel):
    """Represents a DM channel."""

    recipients: list[UserModel] | None
    """Recipients of the DM."""

    icon: str | None
    """Icon hash of the group DM."""

    owner_id: Snowflake | None
    """ID of the creator of the group DM."""

    application_id: Snowflake | None
    """ID of the application that created the DM."""



from dataclasses import dataclass, field

from ...core.model import DataModel

from ...enums.command import CommandType

@dataclass
class UserCommandPart(DataModel):
    """Represents the user command object."""

    name: str | None = None
    """Name of the command."""

    type: CommandType = field(init=False, default=CommandType.USER)
    """Command type. Always `CommandType.USER` for this class."""

@dataclass
class MessageCommandPart(DataModel):
    """Represents the message command object."""
    
    name: str | None = None
    """Name of the command."""

    type: CommandType = field(init=False, default=CommandType.MESSAGE)
    """Command type. Always `CommandType.MESSAGE` for this class."""



from scurrypy.api.interactions import MessageComponentDataModel, ModalDataModel

from ..interactions.ctx import InteractionContext

class MessageComponentContext(InteractionContext):
    data: MessageComponentDataModel

class ComponentModalContext(InteractionContext):
    data: ModalDataModel



from typing import TypedDict

from ..enums.message import MessageFlags

from ..api.components.layout import ActionRow, Container
from ..api.messages.attachment import AttachmentPart
from ..api.messages.embed import Embed
from ..api.messages.message import MessageReferencePart

class EditMessageParams(TypedDict, total=False):
    """Parameters for editing a message"""

    content: str
    """Message text content."""

    flags: MessageFlags
    """Message flags.
    
    !!! note
        If omitted, existing message flags are not preserved automatically.
        Passing endpoint flags without explicitly setting `flags`
        will construct a new flags value.
    """

    components: list[ActionRow | Container]
    """Components to be attached to this message."""

    attachments: list[AttachmentPart]
    """Attachments to be attached to this message."""

    embeds: list[Embed]
    """Embeds to be attached to this message."""

    message_reference: MessageReferencePart
    """Message reference if reply."""



from dataclasses import dataclass

from ...core.model import DataModel

from ..components.layout import Label

@dataclass
class ModalPart(DataModel):
    """Represents the Modal object."""

    title: str | None = None
    """Title of the popup modal."""

    custom_id: str | None = None
    """ID for the modal."""

    components: list[Label] | None = None
    """1 to 5 components that make up the modal."""



from dataclasses import dataclass, field

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...bases.channel import GuildChannelCreate

from ...enums.channel import ChannelType

@dataclass
class GuildAnnouncementChannelPart(DataModel, GuildChannelCreate):
    """Parameters for creating a guild announcement channel."""

    name: str | None = None
    """Name of the channel."""

    topic: str | None = None
    """Topic of the channel."""

    position: int | None = None
    """Sorting position of the channel (channels with the same position are sorted by id)."""

    parent_id: Snowflake | None = None
    """Category ID of the channel."""

    nsfw: bool | None = None
    """If the channel is flagged NSFW."""

    default_auto_archive_duration: int | None = None
    """Default duration in minutes threads will be hidden after period of inactivity."""

    default_thread_rate_limit_per_user: int | None = None
    """Rate limit per user set on newly created threads.

    !!! note
        This field does not live update!
    """

    type: ChannelType = field(init=False, default=ChannelType.GUILD_ANNOUNCEMENT)
    """Type of channel. Always `ChannelType.GUILD_ANNOUNCEMENT` for this class."""



# scurrypy/etc/commands

from .addon import CommandsAddon
from .ctx import ApplicationCommandContext, AutocompleteApplicationCommandContext

__all__ = [
    "CommandsAddon",
    
    "ApplicationCommandContext",
    "AutocompleteApplicationCommandContext"
]



from .enum_types import DiscordString, DiscordTypes

class PromptType(DiscordTypes):
    MULTIPLE_CHOICE = 0
    DROPDOWN = 1

class OnboardingMode(DiscordTypes):
    ONBOARDING_DEFAULT = 0
    ONBOARDING_ADVANCED = 1

class StickerType(DiscordTypes):
    """Sticker types."""

    STANDARD = 1
    """An official sticker in a pack."""

    GUILD = 2
    """A sticker uploaded to a guild for the guild's members."""

class StickerFormatType(DiscordTypes):
    PNG = 1
    APNG = 2
    LOTTIE = 3
    GIF = 4

class GuildFeature(DiscordString):
    NEWS = "NEWS"
    ROLE_ICONS = "ROLE_ICONS"
    ANIMATED_ICON = "ANIMATED_ICON"
    INVITE_SPLASH = "INVITE_SPLASH"
    DISCOVERABLE = "DISCOVERABLE"
    BANNER = "BANNER"
    ANIMATED_BANNER = "ANIMATED_BANNER"
    PARTNERED = "PARTNERED"



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...enums.guild import OnboardingMode, PromptType

from ..emoji import EmojiModel

@dataclass
class OnboardingPromptOptionModel(DataModel):
    """Represents a guild's prompt option for onboarding."""

    id: Snowflake
    """ID of the prompt option."""

    channel_ids: list[Snowflake]
    """Channel IDs a member is added to when selected."""

    role_ids: list[Snowflake]
    """Role IDs a member is given when selected."""

    emoji: EmojiModel | None
    """Emoji for the option."""

    emoji_id: Snowflake | None
    """ID for the emoji of the option."""

    emoji_name: str | None
    """Name for the emoji of the option."""

    emoji_animated: bool | None
    """Whether the emoji of the option is animated."""

    title: str
    """Title of the option."""

    description: str
    """Description of the option."""

@dataclass
class OnboardingPromptModel(DataModel):
    """Represents a guild's prompt for onboarding."""

    id: Snowflake
    """ID of the prompt."""

    type: PromptType
    """Type of prompt."""

    options: list[OnboardingPromptOptionModel]
    """Options available with the prompt."""

    title: str
    """Title of the prompt."""

    single_select: bool
    """Whether users are limited to selecting one option."""

    required: bool
    """Whether the prompt is required for completing the onboarding process."""

    in_onboarding: bool
    """Whether the prompt is present in the onboarding flow."""

@dataclass
class GuildOnboadingModel(DataModel):
    """Represents a guild's onboarding flow."""

    guild_id: Snowflake
    """ID of the guild for onboarding."""

    prompts: list[OnboardingPromptModel]
    """Prompts shown during onboarding."""

    default_channel_ids: list[Snowflake]
    """Channel IDs members are opted into by default."""

    enabled: bool
    """Whether onboarding is enabled for the guild."""

    mode: OnboardingMode
    """Current mode of onboarding."""

@dataclass
class OnboardingPromptOptionPart(DataModel):
    """Represents fields for creating an onboarding prompt option."""

    channel_ids: list[Snowflake] | None = None
    """	IDs for channels a member is added to when the option is selected."""

    role_ids: list[Snowflake] | None = None
    """IDs for roles assigned to a member when the option is selected."""

    wmoji_id: Snowflake | None = None
    """	Emoji ID of the option."""

    emoji_name: str | None = None
    """Emoji name of the option."""

    emoji_animated: bool | None = None
    """Whether the emoji is animated."""

    title: str | None = None
    """Title of the option."""

    description: str | None = None
    """Description of the option."""

@dataclass
class OnboardingPromptPart(DataModel):
    """Represents fields for creating an onboarding prompt."""

    type: PromptType | None = None
    """Type of prompt."""

    options: list[OnboardingPromptOptionPart] | None = None
    """Options available with the prompt."""

    title: str | None = None
    """Title of the prompt."""

    single_select: bool | None = None
    """Whether the users are limited to selecting one option."""

    required: bool | None = None
    """Whether the prompt is required to complete the onboarding flow."""



from dataclasses import dataclass
from typing import Unpack

from .base_resource import BaseResource

from ..core.snowflake import Snowflake
from ..core.serialization import serialize

from ..api.messages.message import MessageModel, PinnedMessageModel, MessagePart
from ..api.channels.channel import ChannelModel
from ..api.channels.followed import FollowedChannelModel
from ..api.channels.threads import ThreadMemberModel, ArchivedThreadsModel, ThreadFromMessagePart, ThreadWithoutMessagePart

from ..api.invite import InviteModel, InvitePart, InviteWithMetadataModel

from ..params.channel import EditGuildChannelParams, EditThreadChannelParams

@dataclass
class Channel(BaseResource):
    """Represents a Discord channel."""

    id: Snowflake
    """ID of the channel."""

    # --- CHANNEL ---
    async def fetch(self) -> ChannelModel:
        """Fetch the full channel data from Discord.

        Returns:
            (ChannelModel): queried channel
        """
        data = await self.http.request("GET", f"/channels/{self.id}")

        return ChannelModel.from_dict(data)

    async def delete(self) -> None:
        """Deletes this channel from the server. 
        Fires [`ChannelUpdateEvent`][scurrypy.events.channel_events.ChannelUpdateEvent] if success,
        and [`ChannelDeleteEvent`][scurrypy.events.channel_events.ChannelDeleteEvent] 
            (or [`ThreadDeleteEvent`][scurrypy.events.thread_events.ThreadDeleteEvent] if a thread).

        !!! important "Permissions"
            Requires `MANAGE_CHANNELS` and `MANAGE_THREADS`
        """
        await self.http.request("DELETE", f"/channels/{self.id}")

    async def follow(self, webhook_channel_id: Snowflake) -> FollowedChannelModel:
        """Follow announcement channel to send messages to a target channel.
        Fires [`WebhooksUpdateEvent`][scurrypy.events.channel_events.WebhooksUpdateEvent].

        Args:
            webhook_channel_id (Snowflake): ID of target channel

        Returns:
            (FollowedChannelModel): followed channel
        """
        data = await self.http.request(
            'POST', 
            f'/channels/{self.id}/followers', 
            params={'webhook_channel_id': webhook_channel_id}
        )

        return FollowedChannelModel.from_dict(data)

    # --- GUILD CHANNEL ---
    async def edit_guild_channel(self, **options: Unpack[EditGuildChannelParams]) -> ChannelModel:
        """Edit this channel. 
        Fires [`ChannelUpdateEvent`][scurrypy.events.channel_events.ChannelUpdateEvent].
        
        !!! note
            If modifying a category, all child channels also fire [`ChannelUpdateEvent`][scurrypy.events.channel_events.ChannelUpdateEvent].

        !!! important "Permissions"
            Requires `MANAGE_CHANNELS`

        Args:
            options (EditGuildChannelParams): channel fields to edit

        Returns:
            (ChannelModel): updated channel
        """
        options = serialize(options)

        data = await self.http.request('PATCH', f'/channels/{self.id}', data=options)

        return ChannelModel.from_dict(data)
    
    # --- MESSAGES ---
    async def fetch_messages(self, limit: int = 50, before: Snowflake = None, after: Snowflake = None, around: Snowflake = None) -> list[MessageModel]:
        """Fetches this channel's messages.

        !!! important "Permissions"
            Requires `VIEW_CHANNEL` and `READ_MESSAGE_HISTORY`

        Args:
            limit (int, optional): Max number of messages to return. Range 1 - 100. Defaults to `50`.
            before (Snowflake, optional): get messages before this message ID
            after (Snowflake, optional): get messages after this message ID
            around (Snowflake, optional): get messages around this message ID

        Returns:
            (list[MessageModel]): queried list of messages
        """
        params = {
            "limit": limit,
            "before": before,
            "after": after,
            "around": around
        }

        data = await self.http.request('GET', f'/channels/{self.id}/messages', params=params)

        return [MessageModel.from_dict(msg) for msg in data]

    async def fetch_pins(self, limit: int = 50, before: str = None) -> list[PinnedMessageModel]:
        """Get this channel's pinned messages.

        !!! important "Permissions"
            Requires `VIEW_CHANNEL` and `READ_MESSAGE_HISTORY`
            
        !!! note
            * Creates a `PUBLIC_THREAD` when called on a `GUILD_TEXT` channel
            * Creates an `ANNOUNCEMENT_THREAD` when called on a `GUILD_ANNOUNCEMENT` channel
        
        !!! warning
            Does not work on a `GUILD_FORUM` channel!

        Args:
            before (str, optional): get pinned messages before this ISO8601 timestamp
            limit (int, optional): Max number of pinned messages to return. Range 1 - 50. Defaults to `50`.
        
        Returns:
            (list[PinnedMessageModel]): queried list of pinned messages
        """
        # Set default limit if user didn't supply one
        params = {
            "limit": limit,
            "before": before
        }

        data = await self.http.request('GET', f'/channels/{self.id}/pins', params=params)

        return [PinnedMessageModel.from_dict(item) for item in data]

    async def send(self, message: str | MessagePart) -> MessageModel:
        """Send a message to this channel.
        Fires [`MessageCreateEvent`][scurrypy.events.message_events.MessageCreateEvent].

        !!! important "Permissions"
            Requires `SEND_MESSAGES`

        Args:
            message (str | MessagePart): content as a string or MessagePart

        Returns:
            (MessageModel): created message
        """
        if isinstance(message, str):
            message = MessagePart(content=message)

        if message.attachments:
            message = message._prepare()
        
        data = await self.http.request(
            "POST", 
            f"/channels/{self.id}/messages", 
            data=message.to_dict(),
            files=[fp.path for fp in message.attachments] if message.attachments else None
        )

        return MessageModel.from_dict(data)

    async def bulk_delete_messages(self, message_ids: list[Snowflake]) -> None:
        """Delete multiple messages in a single request.
        Fires [`BulkMessageDeleteEvent`][scurrypy.events.message_events.BulkMessageDeleteEvent].
        
        !!! important "Permissions"
            Requires `MANAGE_MESSAGES`.

        !!! important
            Messages **older than 2 weeks** will fail to get deleted!

        !!! note
            Only available for `GUILD_TEXT` channels.

        Args:
            message_ids (list[Snowflake]): IDs of the messages to delete range(2, 100)
        """
        await self.http.request(
            'POST', 
            f'/channels/{self.id}/messages/bulk-delete', 
            data={'messages': message_ids}
        )

    # --- INVITES ---
    async def fetch_invites(self) -> list[InviteWithMetadataModel]:
        """Fetch a list of invites for this channel.

        !!! important "Permissions"
            Requires `MANAGE_CHANNELS`

        !!! note
            Only usable on guild channels.

        Returns:
            list[InviteWithMetadataModel]: queried list of invites
        """
        data = await self.http.request('GET', f'/channels/{self.id}/invites')

        return [InviteWithMetadataModel.from_dict(i) for i in data]

    async def create_invite(self, invite: InvitePart) -> InviteModel:
        """Create a new invite for this channel.
        Fires [`InviteCreateEvent`][scurrypy.events.invite_events.InviteCreateEvent].

        !!! important "Permissions"
            Requires `CREATE_INSTANT_INVITE`

        Args:
            invite (InvitePart): invite to create

        Returns:
            (InviteModel): created invite object 
        """
        data = await self.http.request('POST', f'/channels/{self.id}/invites', data=invite.to_dict())

        return InviteModel.from_dict(data)

    # --- THREAD CHANNELS ---
    async def fetch_thread_member(self, user_id: Snowflake, with_member: bool = False) -> ThreadMemberModel:
        """Fetch a thread emmber of the specified user ID from this thread.

        Args:
            user_id (Snowflake): ID of the user to fetch
            with_member (bool, optional): whether to include the member object. Defaults to `False`.
        
        Returns:
            (ThreadMemberModel): queried thread member
        """

        params = { 'with_member': with_member }

        data = await self.http.request('GET', f'/channels/{self.id}/thread-members/{user_id}', params=params)

        return ThreadMemberModel.from_dict(data)
    
    async def fetch_thread_members(self, limit: int = 100, after: Snowflake = None, with_member: bool = False) -> list[ThreadMemberModel]:
        """Fetch all members of this thread.

        !!! warning
            Requires the `GUILD_MEMBERS` privileged intent to use!

        Args:
            limit (int, optional): Max number of thread members to return. Range 0 - 100. Defaults to `100`.
            after (Snowflake, optional): members after this user ID
            with_member (bool, optional): whether to include the member object. Defaults to `False`.

        Returns:
            (list[ThreadMemberModel]): queried list of thread members
        """

        params = {
            'with_member': with_member,
            'after': after,
            'limit': limit
        }

        data = await self.http.request('GET', f"/channels/{self.id}/thread-members", params=params)

        return [ThreadMemberModel.from_dict(n) for n in data]

    async def create_thread_from_message(self, message_id: Snowflake, thread: ThreadFromMessagePart) -> ChannelModel:
        """Create a thread from a message (attached to the message). 
        Fires [`ThreadCreateEvent`][scurrypy.events.thread_events.ThreadCreateEvent] 
        and [`MessageUpdateEvent`][scurrypy.events.message_events.MessageUpdateEvent].

        Args:
            message_id (Snowflake): ID of the message to attach the thread
            thread (ThreadFromMessagePart): thread to attach

        Returns:
            ChannelModel: new thread
        """

        data = await self.http.request('POST', f"channels/{self.id}/messages/{message_id}/threads", data=thread.to_dict())

        return ChannelModel.from_dict(data)

    async def create_thread_without_message(self, thread: ThreadWithoutMessagePart) -> ChannelModel:
        """Create a thread not connected to an existing message.
        Fires [`ThreadCreateEvent`][scurrypy.events.thread_events.ThreadCreateEvent].

        Args:
            thread (ThreadWithoutMessagePart): thread to create

        Returns:
            ChannelModel: new thread
        """

        data = await self.http.request('POST', f'/channels/{self.id}/threads', data=thread.to_dict())

        return ChannelModel.from_dict(data)

    async def edit_thread(self, **options: Unpack[EditThreadChannelParams]) -> ChannelModel:
        """Edit this thread. 
        Fires [`ChannelUpdateEvent`][scurrypy.events.channel_events.ChannelUpdateEvent].

        !!! important "Permissions"
            Requires `MANAGE_CHANNELS`

        !!! important
            Requires `archived` be `False` or set to `False`.

        Args:
            options (EditThreadChannelParams): channel fields to edit

        Returns:
            (ChannelModel): updated channel
        """

        data = await self.http.request('PATCH', f'/channels/{self.id}', data=options)

        return ChannelModel.from_dict(data)

    async def join_thread(self) -> None:
        """Add the bot to this thread.
        Fires [`ThreadMembersUpdateEvent`][scurrypy.events.thread_events.ThreadMembersUpdateEvent] 
        and [`ThreadCreateEvent`][scurrypy.events.thread_events.ThreadCreateEvent].

        !!! important
            Required the thread NOT be archived.
        """
        await self.http.request('PUT', f'/channels/{self.id}/thread-members/@me')

    async def leave_thread(self) -> None:
        """Remove the bot from a thread.
        Fires [`ThreadMembersUpdateEvent`][scurrypy.events.thread_events.ThreadMembersUpdateEvent].

        !!! important
            Required the thread NOT be archived.
        """
        await self.http.request('DELETE', f'/channels/{self.id}/thread-members/@me')

    async def add_thread_member(self, user_id: Snowflake) -> None:
        """Add a user to this thread.
        Fires [`ThreadMembersUpdateEvent`][scurrypy.events.thread_events.ThreadMembersUpdateEvent].

        Args:
            user_id (Snowflake): ID of the user to add
        """
        await self.http.request('PUT', f'/channels/{self.id}/thread-members/{user_id}')

    async def remove_thread_member(self, user_id: Snowflake) -> None:
        """Remove a user to this thread.
        Fires [`ThreadMembersUpdateEvent`][scurrypy.events.thread_events.ThreadMembersUpdateEvent].

        Args:
            user_id (Snowflake): ID of the user to remove
        """
        await self.http.request('DELETE', f'/channels/{self.id}/thread-members/{user_id}')

    async def fetch_public_archived_threads(self, before: str = None, limit: int = None) -> ArchivedThreadsModel:
        """Fetch archived public threads in this channel.

        !!! important "Permissions"
            Requires `READ_MESSAGE_HISTORY`

        !!! note:
            Returns `PUBLIC_THREAD` threads if this is a `GUILD_TEXT` channel.
            Returns `ANNOUNCEMENT_THREAD` if this is a `GUILD_ANNOUNCEMENT` channel.

        !!! note
            Threads are ordered by `archive_timestamp` in descending order.

        Args:
            before (str, optional): threads archived before this timestamp
            limit (int, optional): max numer of threads to fetch

        Returns:
            (ArchivedThreadsModel): queried public archived threads
        """
        data = await self.http.request(
            'GET', 
            f'/channels/{self.id}/threads/archived/public', 
            params={
                'before': before, 
                'limit': limit
            }
        )

        return ArchivedThreadsModel.from_dict(data)
    
    async def fetch_private_archived_threads(self, before: str = None, limit: int = None) -> ArchivedThreadsModel:
        """Fetch archived private threads in this channel.

        !!! important "Permissions"
            Requires `READ_MESSAGE_HISTORY` and `MANAGE_THREADS`

        !!! note
            Threads are ordered by `archive_timestamp` in descending order.

        Args:
            before (str, optional): threads archived before this timestamp
            limit (int, optional): max numer of threads to fetch

        Returns:
            (ArchivedThreadsModel): queried private archived threads
        """
        data = await self.http.request(
            'GET', 
            f'/channels/{self.id}/threads/archived/private', 
            params={
                'before': before, 
                'limit': limit
            }
        )

        return ArchivedThreadsModel.from_dict(data)
    
    async def fetch_joined_private_archived_threads(self, before: str = None, limit: int = None) -> ArchivedThreadsModel:
        """Fetch archived private threads in this channel the bot has joined.

        !!! important "Permissions"
            Requires `READ_MESSAGE_HISTORY`

        !!! note
            Threads are ordered by their `id` in descending order.

        Args:
            before (str, optional): threads archived before this timestamp
            limit (int, optional): max numer of threads to fetch

        Returns:
            (ArchivedThreadsModel): queried private archived threads
        """
        data = await self.http.request(
            'GET', 
            f'/channels/{self.id}/users/@me/threads/archived/private', 
            params={
                'before': before, 
                'limit': limit
            }
        )

        return ArchivedThreadsModel.from_dict(data)



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

@dataclass
class TagModel(DataModel):
    """Represents the tag object found in `GUILD_FORUM` channels."""
    
    id: Snowflake
    """ID of the tag."""

    name: str
    """Name of the tag."""

    moderated: bool
    """Whether the tag can only be added/removed by a member with `MANAGE_THREADS`."""
    
    emoji_id: Snowflake
    """ID of a guild's custom emoji."""
    
    emoji_name: str
    """Unicode character of the emoji."""

@dataclass
class TagPart(DataModel):
    """Represents the tag object found in `GUILD_FORUM` channels."""
    
    name: str = None
    """Name of the tag."""

    moderated: bool = None
    """Whether the tag can only be added/removed by a member with `MANAGE_THREADS`."""
    
    emoji_id: Snowflake = None
    """ID of a guild's custom emoji."""
    
    emoji_name: str = None
    """Unicode character of the emoji."""



# scurrypy/resources

from .application import Application
from .channel import Channel
from .command import (
    GuildCommand,
    GlobalCommand
)
from .emoji import (
    ApplicationEmoji,
    GuildEmoji
)
from .guild import Guild
from .interaction import Interaction
from .invite import Invite
from .message import Message
from .sticker import Sticker
from .user import User

__all__ = [
    "Application",

    "Channel",

    "GuildCommand", 
    "GlobalCommand",

    "ApplicationEmoji",
    "GuildEmoji",

    "Guild",

    "Interaction",

    "Invite",

    "Message",

    "Sticker",
    
    "User"
]



from scurrypy import Client
from scurrypy.resources import Channel
from scurrypy.events import MessageCreateEvent

class PrefixCommandContext(Channel):
    """Useful message prefix event info."""

    def __init__(self, bot: Client, event: MessageCreateEvent, args: list[str]):
        super().__init__(bot.http, event.channel_id)
        self.bot = bot
        self.event = event
        self.args = args

    @property
    def author(self):
        """Author of the prefix command."""
        return self.event.author

    @property
    def guild(self):
        """Guild resource of the prefix command."""
        return self.bot.guild(self.event.guild_id)

    @property
    def channel(self):
        """Channel resource of the prefix command."""
        return self.bot.channel(self.event.channel_id)

    @property
    def message(self):
        """Message resource of the prefix command."""
        return self.bot.message(self.event.channel_id, self.event.id)



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from .base_event import Event

from ..enums.channel import ChannelType

from ..api.channels.channel import ChannelModel
from ..api.channels.threads import ThreadMemberModel

@dataclass
class ThreadCreateEvent(Event, ChannelModel):
    """Received when a thread is created."""

    newly_created: bool
    """Whether the thread has just been created."""

@dataclass
class ThreadUpdateEvent(Event, ChannelModel):
    """Received when a thread is updated.
    
    !!! note
        Not send when `last_message_id` is changed.
    """
    pass

@dataclass
class ThreadDeleteEvent(Event, DataModel):
    """Received when a thread is deleted."""

    id: Snowflake
    """ID of the thread."""
    
    guild_id: Snowflake | None
    """Guild ID of the thread."""
    
    parent_id: Snowflake
    """ID of the parent channel."""
    
    type: ChannelType
    """Type of thread."""

@dataclass
class ThreadMemberUpdateEvent(Event, ThreadMemberModel):
    """Received when a thread member for the bot is updated."""

    guild_id: Snowflake
    """ID of the guild."""

@dataclass
class ThreadMembersUpdateEvent(Event, DataModel):
    """Received when someone is added or removed from a thread.
    
    !!! important
        Without the `GUILD_MEMBERS` privileged intent, this event only fires if the 
        bot was added or removed from a thread.
    """

    id: Snowflake
    """ID of the thread."""

    guild_id: Snowflake
    """ID of the guild."""

    member_count: int
    """Approximate number of members in the thread (max `50`)."""

    added_members: list[ThreadMemberModel] | None
    """Users who were added to the thread"""

    removed_member_ids: list[Snowflake] | None
    """ID of the users who were removed from the thread."""

@dataclass
class ThreadListSyncEvent(Event, DataModel):
    """Received when the bot gains access to a channel."""

    guild_id: Snowflake
    """ID of the guild."""

    channel_ids: list[Snowflake] | None
    """Parent channel IDs of the threads being synced."""

    threads: list[ChannelModel]
    """Active threads in the given channel that the bot can access."""

    members: list[ThreadMemberModel]
    """Thread members from the synced threads that the bot an access."""



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

@dataclass
class DefaultReactionModel(DataModel):
    """Represents the default reaction for a `GUILD_FORUM` post."""

    emoji_id: Snowflake
    """ID of the guild's custom emoji."""

    emoji_name: str
    """Unicode character of the emoji."""

@dataclass
class DefaultReactionPart(DataModel):
    """Represents the default reaction for a `GUILD_FORUM` post."""

    emoji_id: int | None = None
    """ID of the guild's custom emoji."""

    emoji_name: str | None = None
    """Unicode character of the emoji."""



class GuildChannelCreate:
    """Marker base for all guild channel creation payloads.
    
    !!! tip "Variants"
        [`GuildTextChannelPart`][scurrypy.api.channels.GuildTextChannelPart]
        [`GuildAnnouncementChannelPart`][scurrypy.api.channels.GuildAnnouncementChannelPart]
        [`GuildForumChannelPart`][scurrypy.api.channels.GuildForumChannelPart]
    """
    __slots__ = ()



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from .base_event import Event

from ..api.channels.channel import ChannelModel

@dataclass
class ChannelCreateEvent(Event, ChannelModel):
    """Received when a guild channel has been created."""
    pass

@dataclass
class ChannelUpdateEvent(Event, ChannelModel):
    """Received when a guild channel has been updated.

    !!! note
        Not send when `last_message_id` is changed.
    """
    pass

@dataclass
class ChannelDeleteEvent(Event, ChannelModel):
    """Received when a guild channel has been deleted."""
    pass

@dataclass
class ChannelPinsUpdateEvent(Event, DataModel):
    """Pin update event."""
    
    channel_id: Snowflake
    """ID of channel where the pins were updated."""

    guild_id: Snowflake | None
    """ID of the guild where the pins were updated."""

    last_pin_timestamp: str | None
    """ISO8601 formatted timestamp of the last pinned message in the channel."""

@dataclass
class WebhooksUpdateEvent(Event, DataModel):
    """Received when a guild's channel webhook is created, updated, or deleted."""

    guild_id: Snowflake
    """ID of the guild."""

    channel_id: Snowflake
    """ID of the channel."""



from dataclasses import dataclass

from ...bases.components import Component

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...enums.message import MessageType, MessageFlags, MessageReferenceType

from ..components.layout import Container, ActionRow
from ..messages.attachment import AttachmentModel, AttachmentPart
from ..messages.embed import Embed
from ..messages.reaction import ReactionModel
from ..channels.channel import ChannelModel
from ..guilds.role import GuildRoleModel

from ..user import UserModel

@dataclass
class MessageModel(DataModel):
    """Represents a Discord message."""

    id: Snowflake
    """ID of the message."""

    channel_id: Snowflake
    """Channel ID of the message."""

    author: UserModel
    """User data of author of the message."""
    
    content: str
    """Content of the message."""

    pinned: bool
    """If the message is pinned."""

    type: MessageType
    """Type of message."""

    flags: MessageFlags
    """Message flags."""

    attachments: list[AttachmentModel]
    """Attached files."""

    embeds: list[Embed]
    """Embedded content."""

    thread: ChannelModel | None
    """Thread created from the message."""

    reactions: list[ReactionModel]
    """Reactions to the message."""

    webhook_id: Snowflake | None
    """ID of the webhook if the message is a webhook."""

    timestamp: str | None
    """Timestamp of when the message was sent."""

    edited_timestamp: str | None
    """Timestamp of when the message was last edited."""

    mention_everyone: bool
    """Whether the message mentions everyone."""

    mentions: list[UserModel]
    """List of mentioned users in the message."""

    mention_roles: list[GuildRoleModel]
    """List of mentioned roles in the message."""

    components: list[Component]
    """Components contained in the message."""

@dataclass
class PinnedMessageModel(DataModel):
    """Represents a pinned message."""

    message: MessageModel
    """Message resource of the pinned message."""

    pinned_at: str | None
    """ISO8601 timestamp of when the message was pinned."""

@dataclass
class MessageReferencePart(DataModel):
    """Represents the Message Reference object."""

    message_id: Snowflake | None = None
    """ID of the originating message."""

    channel_id: Snowflake | None = None
    """
        Channel ID of the originating message.
        !!! note
            Optional for default type, but REQUIRED for forwards.
    """

    type: MessageReferenceType | None = None
    """Type of reference. Discord defaults to `MessageReferenceTypes.DEFAULT`."""

@dataclass
class MessagePart(DataModel):
    """Represents a Discord Message."""

    content: str | None = None
    """Message text content."""

    flags: MessageFlags = None
    """Message flags. Discord defaults to `MessageFlags.NO_FLAGS`."""

    components: list[ActionRow | Container] | None = None
    """Components to be attached to this message."""

    attachments: list[AttachmentPart] | None = None
    """Attachments to be attached to this message."""

    embeds: list[Embed] | None = None
    """Embeds to be attached to this message."""

    message_reference: MessageReferencePart | None = None
    """Message reference if reply."""

    def _prepare(self):
        """Prepares MessagePart for ANY internally set attributes.

        Returns:
            (MessagePart): self
        """
        # set attachment IDs (if any)
        if self.attachments:
            for idx, file in enumerate(self.attachments):
                file.id = idx
        
        return self

    def to_dict(self):
        if self.components:
            for component in self.components:
                if not isinstance(component, Container):
                    continue

                if not self.flags or MessageFlags.IS_COMPONENTS_V2 not in self.flags:
                    raise ValueError("V2 components are used but MessageFlags.IS_COMPONENTS_V2 is not set.")
        
        return super().to_dict()



class Component:
    """Marker class for all interaction components and containers."""
    ...

class ActionRowChild: 
    """Marker class for all components that go into an action row.

    !!! tip "Children"
        [`Button`][scurrypy.api.components.Button], 
        [`StringSelect`][scurrypy.api.components.StringSelect], 
        [`UserSelect`][scurrypy.api.components.UserSelect], 
        [`RoleSelect`][scurrypy.api.components.RoleSelect], 
        [`MentionableSelect`][scurrypy.api.components.MentionableSelect], 
        [`ChannelSelect`][scurrypy.api.components.ChannelSelect]
    """
    __slots__ = ()

class SectionChild: 
    """Marker class for all components that go into a section.

    !!! tip "Children"
        [`TextDisplay`][scurrypy.api.components.TextDisplay]
    """
    __slots__ = ()

class SectionAccessoryChild: 
    """Marker class for all components that go into a section accessory.
    
    !!! tip "Children"
        [`Button`][scurrypy.api.components.Button], 
        [`Thumbnail`][scurrypy.api.components.Thumbnail]
    """
    __slots__ = ()

class ContainerChild: 
    """Marker class for all components that go into a container.
    
    !!! tip "Children"
        [`ActionRow`][scurrypy.api.components.ActionRow], 
        [`TextDisplay`][scurrypy.api.components.TextDisplay], 
        [`Section`][scurrypy.api.components.Section], 
        [`MediaGallery`][scurrypy.api.components.MediaGallery], 
        [`Separator`][scurrypy.api.components.Separator], 
        [`File`][scurrypy.api.components.File]
    """
    __slots__ = ()

class LabelChild: 
    """Marker class for all components that go into a label.
    
    !!! tip "Children"
        [`TextInput`][scurrypy.api.components.TextInput], 
        [`StringSelect`][scurrypy.api.components.StringSelect], 
        [`UserSelect`][scurrypy.api.components.UserSelect], 
        [`RoleSelect`][scurrypy.api.components.RoleSelect], 
        [`MentionableSelect`][scurrypy.api.components.MentionableSelect], 
        [`ChannelSelect`][scurrypy.api.components.ChannelSelect], 
        [`FileUpload`][scurrypy.api.components.FileUpload]
        [`RadioGroup`][scurrypy.api.components.RadioGroup]
        [`CheckboxGroup`][scurrypy.api.components.CheckboxGroup]
        [`Checkbox`][scurrypy.api.components.Checkbox]
    """
    __slots__ = ()



from dataclasses import dataclass
from typing import Unpack

from .base_resource import BaseResource

from ..core.snowflake import Snowflake
from ..core.serialization import serialize

from ..api.commands.application_command import ApplicationCommandModel
from ..api.commands.slash import SlashCommandPart
from ..api.commands.context import UserCommandPart, MessageCommandPart

from ..params.command import EditGlobalCommandParams, EditGuildCommandParams

@dataclass
class GlobalCommand(BaseResource):
    """Represents a global command."""

    application_id: Snowflake
    """Application ID of the commands."""

    async def fetch(self, command_id: Snowflake) -> ApplicationCommandModel:
        """Fetches a command object.

        Args:
            command_id (int): ID of the command to fetch

        Returns:
            (ApplicationCommandModel): queried application command
        """
        data = await self.http.request('GET', f"applications/{self.application_id}/commands/{command_id}")

        return ApplicationCommandModel.from_dict(data)
    
    async def fetch_all(self) -> list[ApplicationCommandModel]:
        """Fetches ALL global commands.

        Returns:
            (list[ApplicationCommandModel]): queried list of application commands
        """
        data = await self.http.request('GET', f"applications/{self.application_id}/commands")

        return [ApplicationCommandModel.from_dict(cmd) for cmd in data]

    async def create(self, command: SlashCommandPart | UserCommandPart | MessageCommandPart) -> ApplicationCommandModel:
        """Add a command to the client.

        !!! danger
            Creating a command with the same name as an existing command in the same scope will overwrite the old command.

        Args:
            command (SlashCommandPart | UserCommandPart | MessageCommandPart): command to register

        Returns:
            (ApplicationCommandModel): created command
        """
        data = await self.http.request('POST', f"applications/{self.application_id}/commands", data=command.to_dict())

        return ApplicationCommandModel.from_dict(data)

    async def edit(self, command_id: Snowflake, **options: Unpack[EditGlobalCommandParams]) -> ApplicationCommandModel:
        """Edit a command.

        Args:
            command_id (Snowflake): ID of command to edit
            options (EditGlobalCommandParams): command fields to edit

        Returns:
            (ApplicationCommandModel): updated application command
        """
        options = serialize(options)

        data = await self.http.request('PATCH', f"applications/{self.application_id}/commands/{command_id}", data=options)

        return ApplicationCommandModel.from_dict(data)

    async def delete(self, command_id: Snowflake) -> None:
        """Delete a command.

        Args:
            command_id (Snowflake): ID of the command to delete
        """
        await self.http.request('DELETE', f"applications/{self.application_id}/commands/{command_id}")

    async def bulk_overwrite(self, commands: list[SlashCommandPart | UserCommandPart | MessageCommandPart]) -> list[ApplicationCommandModel]:
        """Takes a list of application commands, overwriting existing commands list for this application. 
        
        !!! warning
            Commands that do not already exist will count toward daily application command create limits.

        !!! danger
            This will overwrite all types of application commands: slash commands, user commands, and message commands.

        Args:
            commands (list[SlashCommandPart | UserCommandPart | MessageCommandPart]): commands to register

        Returns:
            (list[ApplicationCommandModel]): created application commands
        """

        data = await self.http.request(
            'PUT', 
            f"applications/{self.application_id}/commands", 
            data=[cmd.to_dict() for cmd in commands]
        )

        return [ApplicationCommandModel.from_dict(cmd) for cmd in data]


@dataclass
class GuildCommand(BaseResource):
    """Represents a guild command."""

    application_id: Snowflake
    """Application ID of the commands."""

    guild_id: Snowflake
    "Guild ID of command."

    async def fetch(self, command_id: Snowflake) -> ApplicationCommandModel:
        """Fetches the command object.

        Args:
            command_id (int): ID of command to fetch

        Returns:
            (ApplicationCommandModel): queried application command
        """
        data = await self.http.request('GET', f"applications/{self.application_id}/guilds/{self.guild_id}/commands/{command_id}")

        return ApplicationCommandModel.from_dict(data)
    
    async def fetch_all(self) -> list[ApplicationCommandModel]:
        """Fetches ALL guild commands.

        Returns:
            (list[ApplicationCommandModel]): queried list of application commands
        """
        data = await self.http.request('GET', f"applications/{self.application_id}/guilds/{self.guild_id}/commands" )

        return [ApplicationCommandModel.from_dict(cmd) for cmd in data]

    async def create(self, command: SlashCommandPart | UserCommandPart | MessageCommandPart) -> ApplicationCommandModel:
        """Add a command to the client.

        !!! danger
            Creating a command with the same name as an existing command in the same scope will overwrite the old command.

        Args:
            command (SlashCommandPart | UserCommandPart | MessageCommandPart): command to register

        Returns:
            (ApplicationCommandModel): created command
        """
        data = await self.http.request('POST', f"applications/{self.application_id}/guilds/{self.guild_id}/commands", data=command.to_dict())

        return ApplicationCommandModel.from_dict(data)

    async def edit(self, command_id: Snowflake, **options: Unpack[EditGuildCommandParams]) -> ApplicationCommandModel:
        """Edit a command.

        Args:
            command_id (Snowflake): ID of command to edit
            options (EditGuildCommandParams): command fields to edit

        Returns:
            (ApplicationCommandModel): updated application command
        """
        options = serialize(options)
        
        data = await self.http.request(
            'PATCH', 
            f"applications/{self.application_id}/guilds/{self.guild_id}/commands/{command_id}", 
            data=options
        )

        return ApplicationCommandModel.from_dict(data)

    async def delete(self, command_id: Snowflake) -> None:
        """Delete a command.

        Args:
            command_id (Snowflake): ID of command to delete
        """
        await self.http.request('DELETE', f"applications/{self.application_id}/guilds/{self.guild_id}/commands/{command_id}")

    async def bulk_overwrite(self, commands: list[SlashCommandPart | UserCommandPart | MessageCommandPart]) -> list[ApplicationCommandModel]:
        """Takes a list of application commands, overwriting existing commands list for this guild. 
        
        !!! warning
            Commands that do not already exist will count toward daily application command create limits.

        !!! danger
            This will overwrite all types of application commands: slash commands, user commands, and message commands.

        Args:
            commands (list[SlashCommandPart | UserCommandPart | MessageCommandPart]): commands to register

        Returns:
            (list[ApplicationCommandModel]): created application commands
        """
        data = await self.http.request(
            'PUT', 
            f"applications/{self.application_id}/guilds/{self.guild_id}/commands", 
            data=[cmd.to_dict() for cmd in commands]
        )

        return [ApplicationCommandModel.from_dict(cmd) for cmd in data]



from dataclasses import dataclass, field

from ...core.model import DataModel

from ...bases.components import (
    Component,
    ActionRowChild, 
    SectionAccessoryChild
)

from ...enums.components import (
    ComponentType,
    ButtonStyle
)

from ..emoji import EmojiModel

@dataclass
class Button(DataModel, Component, ActionRowChild, SectionAccessoryChild):
    """Represents the Button component.
    
    A pressable button!
    """

    style: ButtonStyle | None = None
    """A button style."""

    custom_id: str | None = None
    """ID for the button. Do not supply for `ButtonStyles.LINK` style buttons."""

    label: str | None = None
    """Text that appears on the button."""

    emoji: EmojiModel | None = None
    """Emoji icon for the button."""

    url: str | None = None
    """URL for link-style buttons."""

    disabled: bool | None = None
    """Whether the button is disabled. Discord defaults to `False`."""

    link: str | None = None
    """Hyperlink for button. For `ButtonStyles.LINK` style only."""

    type: ComponentType = field(init=False, default=ComponentType.BUTTON)
    """Component type. Always `ComponentType.BUTTON` for this class."""



# scurrypy/api

from .application import ApplicationModel
from .emoji import (
    EmojiModel, 
    ApplicationEmojiPart, 
    GuildEmojiPart
)
from .image_data import (
    ImageDataPart, 
    ImageAssetPart
)
from .integration import IntegrationModel
from .invite import (
    InviteModel, 
    InviteWithMetadataModel, 
    InvitePart
)
from .user import (
    UserModel, 
    GuildMemberModel
)

__all__ = [
    "ApplicationModel",

    "EmojiModel", 
    "ApplicationEmojiPart", 
    "GuildEmojiPart",

    "ImageDataPart", 
    "ImageAssetPart",

    "IntegrationModel",

    "InviteModel", 
    "InviteWithMetadataModel", 
    "InvitePart",

    "UserModel", 
    "GuildMemberModel"
]



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from .base_event import Event

from ..api.messages.message import MessageModel

from ..api.user import GuildMemberModel

@dataclass
class MessageCreateEvent(Event, MessageModel):
    """Received when a message is created.
    
    !!! note
        `member` may be missing on `MESSAGE_CREATE` and `MESSAGE_UPDATE`. Use `author` when you need the user.
    """

    guild_id: Snowflake | None
    """Guild ID of the updated message (if in a guild channel)."""

    member: GuildMemberModel | None
    """Partial Member object of the author of the message."""

@dataclass
class MessageUpdateEvent(Event, MessageModel):
    """Received when a message is updated."""

    guild_id: Snowflake | None
    """Guild ID of the updated message (if in a guild channel)."""

    member: GuildMemberModel | None
    """Partial Member object of the author of the message."""

@dataclass
class MessageDeleteEvent(Event, DataModel):
    """Received when a message is deleted."""

    id: Snowflake
    """ID of the deleted message."""

    channel_id: Snowflake
    """Channel ID of the deleted message."""

    guild_id: Snowflake | None
    """Guild ID of the deleted message (if in a guild channel)."""

@dataclass
class BulkMessageDeleteEvent(Event, DataModel):
    """Received when bulk deleting messages."""

    ids: list[Snowflake]
    """IDs of the messages that were deleted."""

    channel_id: Snowflake
    """ID of the channel."""

    guild_id: Snowflake | None
    """ID of the guild."""



from dataclasses import dataclass

from .base_resource import BaseResource

from ..core.snowflake import Snowflake

from ..api.messages.sticker import StickerModel, StickerPackModel

@dataclass
class Sticker(BaseResource):
    """Represents the Sticker resource."""

    async def fetch(self, sticker_id: Snowflake) -> StickerModel:
        """Fetch a sticker.
        
        Args:
            sticker_id (Snowflake): ID of the sticker to fetch

        Returns:
            (StickerModel): queried sticker
        """
        data = await self.http.request('GET', f'/stickers/{sticker_id}')

        return StickerModel.from_dict(data)

    async def fetch_sticker_pack(self, pack_id: Snowflake) -> StickerPackModel:
        """Fetch a sticker pack.

        Args:
            pack_id (Snowflake): ID of the pack to fetch

        Returns:
            (StickerPackModel): queried sticker pack
        """
        data = await self.http.request('GET', f'/sticker-packs/{pack_id}')

        return StickerPackModel.from_dict(data)

    async def fetch_sticker_packs(self) -> list[StickerPackModel]:
        """Fetch available sticker packs.

        Returns:
            list[StickerPackModel]: queried list of sticker packs.
        """
        data = await self.http.request('GET', '/sticker-packs')

        stickers = data.get('sticker_packs')

        return [StickerPackModel.from_dict(i) for i in stickers]



from dataclasses import dataclass

from ...core.model import DataModel
from ...core.snowflake import Snowflake

from ...enums.channel import ChannelType, AutoArchiveDurationType

from ..user import GuildMemberModel

from .channel import ChannelModel

@dataclass
class ThreadFromMessagePart(DataModel):
    """Parameters for creating a thread attached to a message."""

    name: str = None
    """Name of the thread."""

    rate_limit_per_user: int | None = None
    """Seconds user must wait between sending messages in the channel."""

    auto_archive_duration: AutoArchiveDurationType | None = None
    """Duration in minutes threads will be hidden after period of inactivity."""

@dataclass
class ThreadWithoutMessagePart(DataModel):
    """Parameters for creating a thread without a message."""

    name: str = None
    """Name of the thread."""

    rate_limit_per_user: int | None = None
    """Seconds user must wait between sending messages in the channel."""

    auto_archive_duration: AutoArchiveDurationType | None = None
    """Duration in minutes threads will be hidden after period of inactivity."""

    type: ChannelType | None = None
    """Type of thread to create. If omitted, Discord defaults to `ChannelType.PRIVATE_THREAD`."""

    invitable: bool | None = None
    """Whether non-moderators can add other non-moderators to the thread (private threads only)."""

@dataclass
class ThreadMetadataModel(DataModel):
    """Represents the thread metadata object."""

    archived: bool
    """Whether the thread is archived."""

    auto_archive_duration: int
    """How long to wait until the thread is hidden (in minutes)."""

    archive_timestamp: str
    """ISO8601 timestamp of when the thread's archive status was last changed."""

    locked: bool
    """Whether the thread is locked.
    
    !!! note
        Only users with `MANAGE_THREADS` can unarchive the thread.
    """

    invitable: bool | None
    """Whether non-moderators can add other non-moderators to the thread (private threads only)."""

    create_timestamp: str | None
    """ISO8601 timestamp of thread creation (field only exists after Jan 09, 2022)."""

@dataclass
class ThreadMemberModel(DataModel):
    """Represents a user that has joined a thread."""

    id: Snowflake | None
    """ID of the thread."""

    user_id: Snowflake | None
    """ID of the user."""

    join_timestamp: str
    """ISO8601 timestamp of when the user last joined the thread."""

    member: GuildMemberModel | None
    """Additional information about the user.
    
    !!! note
        Only present when `with_member` is toggled on request.
    """

    default_thread_rate_limit_per_user: int | None
    """Rate limit per user set on newly created threads.
    
    !!! note
        This field does not live update!
    """

@dataclass
class ArchivedThreadsModel(DataModel):
    """Response body for fetching archived threads."""

    threads: list[ChannelModel]
    """The archived threads."""

    members: list[ThreadMemberModel]
    """Thread member for each returned thread the bot has joined."""

    has_more: bool
    """Whether there are additional threads to be returned with subsequent calls."""

@dataclass
class ActiveThreadsModel(DataModel):
    """Response body for fetching active guild threads."""

    threads: list[ChannelModel]
    """The arctive threads."""

    members: list[ThreadMemberModel]
    """Thread member for each returned thread the bot has joined."""

@dataclass
class ThreadChannelModel(ChannelModel):
    """Represents the thread channel."""

    owner_id: Snowflake | None
    """ID of the creator of the thread."""

    application_id: Snowflake | None
    """ID of the application that created thread."""

    thread_metadata: ThreadMetadataModel | None
    """Thread-specific fields not needed by other channels."""

    member: ThreadMemberModel | None
    """Thread member object for the current user if they have joined the thread."""

    default_auto_archive_duration: int | None
    """Default duration in minutes threads will be hidden after period of inactivity."""



# scurrypy/enums

from .application import ApplicationFlags
from .channel import (
    ChannelType,
    ChannelFlags,
    SortOrderType,
    ForumLayoutType,
    AutoArchiveDurationType
)
from .command import (
    CommandType,
    CommandOptionType
)
from .components import (
    ComponentType,
    ButtonStyle,
    SeparatorType,
    TextInputStyle,
    DefaultValueType
)
from .emoji import ReactionType
from .enum_types import (
    DiscordFlags,
    DiscordTypes,
    DiscordString
)
from .events import EventType
from .guild import (
    PromptType,
    OnboardingMode,
    StickerType,
    StickerFormatType,
    GuildFeature
)
from .integration import IntegrationType
from .interaction import (
    InteractionCallbackType,
    InteractionDataType,
    InteractionType
)
from .invite import (
    InviteType
)
from .message import (
    MessageFlags,
    MessageReferenceType,
    MessageType
)
from .permissions import Permissions
from .user import GuildMemberFlags

__all__ = [
    "ApplicationFlags",

    "ChannelType",
    "ChannelFlags",
    "SortOrderType",
    "ForumLayoutType",
    "AutoArchiveDurationType",

    "CommandType",
    "CommandOptionType",

    "ComponentType",
    "ButtonStyle",
    "SeparatorType",
    "TextInputStyle",
    "DefaultValueType",

    "ReactionType",

    "DiscordFlags",
    "DiscordTypes",
    "DiscordString",

    "EventType",

    "PromptType",
    "OnboardingMode",
    "StickerType",
    "StickerFormatType",
    "GuildFeature",

    "IntegrationType",

    "InteractionCallbackType",
    "InteractionDataType",
    "InteractionType",

    "InviteType",

    "MessageFlags",
    "MessageReferenceType",
    "MessageType",

    "Permissions",

    "GuildMemberFlags"
]



from dataclasses import dataclass

from ..core.model import DataModel
from ..core.snowflake import Snowflake

from ..enums.application import ApplicationFlags

from .guilds.guild import GuildModel

from .user import UserModel

@dataclass
class ApplicationModel(DataModel):
    """Represents a Discord application."""

    id: Snowflake
    """ID of the application."""

    name: str
    """Name of the application."""

    icon: str | None
    """Icon hash of the application."""

    description: str | None
    """Description of the application."""

    bot_public: bool | None
    """If the application is public."""

    bot_require_code_grant: bool | None
    """If full OAuth2 code grant is required."""

    bot: UserModel | None
    """Partial bot user object of the application."""

    terms_of_service_url: str | None
    """Terms of Service URL of the application"""

    privacy_policy: str | None
    """Privacy Policy URL of the application."""

    owner: UserModel | None
    """Partial user object of the owner of the application."""

    guild_id: Snowflake | None
    """Guild ID associated with the application."""

    guild: GuildModel | None
    """Partial guild object of the associated guild."""

    cover_image: str | None
    """Image hash of rich presence invite cover."""

    flags: ApplicationFlags | None
    """Public flags of the application."""

    approximate_guild_count: int | None
    """Approximate guild count of the guilds that installed the application."""



from dataclasses import dataclass

from ...core.model import DataModel

from ..user import UserModel

from datetime import datetime, timezone

@dataclass
class EmbedAuthor(DataModel):
    """Represents fields for creating an embed author."""

    name: str = None
    """Name of the author."""

    url: str | None = None
    """URL of the author. http or attachment://<filename> scheme."""

    icon_url: str | None = None
    """URL of author's icon. http or attachment://<filename> scheme."""

@dataclass
class EmbedThumbnail(DataModel):
    """Represents fields for creating an embed thumbnail."""

    url: str | None = None
    """Thumbnail content. http or attachment://<filename> scheme."""

@dataclass
class EmbedField(DataModel):
    """Represents fields for creating an embed field."""

    name: str | None = None
    """Name of the field."""

    value: str | None = None
    """Value of the field."""

    inline: bool | None = False
    """Whether or not this field should display inline. Defaults to `False`."""

@dataclass
class EmbedImage(DataModel):
    """Represents fields for creating an embed image."""

    url: str | None = None
    """Image content. http or attachment://<filename> scheme."""

@dataclass
class EmbedFooter(DataModel):
    """Represents fields for creating an embed footer."""

    text: str | None = None
    """Footer text."""

    icon_url: str | None = None
    """URL of the footer icon. http or attachment://<filename> scheme."""

@dataclass
class Embed(DataModel):
    """Represents fields for creating an embed."""

    title: str | None = None
    """This embed's title."""

    description: str | None = None
    """This embed's description."""

    timestamp: str | None = None
    """Timestamp of when the embed was sent."""

    color: int | None = None
    """Embed's accent color."""

    author: EmbedAuthor | None = None
    """Embed's author."""

    thumbnail: EmbedThumbnail | None = None
    """Embed's thumbnail attachment."""

    image: EmbedImage | None = None
    """Embed's image attachment."""

    fields: list[EmbedField] | None = None
    """List of embed's fields."""

    footer: EmbedFooter | None = None
    """Embed's footer."""

    def set_user_author(self, user: UserModel):
        """Embed author builder.

        Args:
            user (UserModel): user author
        """
        self.author = EmbedAuthor(
            name=user.username,
            icon_url=f"https://cdn.discordapp.com/avatars/{user.id}/{user.avatar}.png"
        )

    def set_timestamp(self, datetime: datetime = None):
        """Embed timestamp builder. Adheres to ISO8601 format.

        Args:
            datetime (datetime, optional): datetime object
        """
        dt = dt or datetime.now(timezone.utc)

        self.timestamp = dt.isoformat()

    def to_dict(self):
        from ..components import Thumbnail as V2Thumbnail

        if isinstance(self.thumbnail, V2Thumbnail):
            raise TypeError(
                "EmbedPart.thumbnail received a ComponentV2 Thumbnail.\n"
                "Use scurrypy.EmbedThumbnail(url) for embed thumbnails."
            )
        
        if isinstance(self.image, V2Thumbnail):
            raise TypeError(
                "EmbedPart.image received a ComponentV2 Thumbnail.\n"
                "Use scurrypy.EmbedImage(url) for embed thumbnails."
            )
        
        return super().to_dict()



from dataclasses import dataclass
from typing import Unpack

from .base_resource import BaseResource

from ..core.snowflake import Snowflake
from ..core.serialization import serialize

from ..enums.message import MessageFlags
from ..enums.interaction import InteractionCallbackType

from ..api.interactions.modal import ModalPart
from ..api.interactions.interaction import InteractionCallbackModel
from ..api.messages.message import MessagePart
from ..api.commands.slash import CommandOptionChoicePart

from ..params.message import EditMessageParams

from .message import _EditMessageMixin

@dataclass
class Interaction(BaseResource, _EditMessageMixin):
    """Represents a Discord Interaction object."""

    id: Snowflake
    """ID of the interaction."""

    token: str
    """Continuation token for responding to the interaction."""

    async def respond(
        self, 
        message: str | MessagePart, 
        *, 
        with_response: bool = False, 
        ephemeral: bool = None, 
        suppress_embeds: bool = None
    ) -> InteractionCallbackModel | None:
        """Create a message in response to an interaction.
        Fires [`InteractionEvent`][scurrypy.events.interaction_events.InteractionEvent]
        and [`MessageCreateEvent`][scurrypy.events.message_events.MessageCreateEvent].

        Args:
            message (str | MessagePart): content as a string or MessagePart
            with_response (bool, optional): if the interaction data should be returned. Defaults to `False`.
            ephemeral (optional, bool): whether the response should be ephemeral
            suppress_embeds (optional, bool): whether the response's embeds should be removed

        Returns:
            (InteractionCallbackModel | None): interaction callback object (if `with_response` is toggled) else None
        """
        if isinstance(message, str):
            message = MessagePart(content=message)

        message.flags = MessageFlags.NO_FLAGS

        if ephemeral:
            message.flags |= MessageFlags.EPHEMERAL

        if suppress_embeds:
            message.flags |= MessageFlags.SUPPRESS_EMBEDS

        content = {
            'type': InteractionCallbackType.CHANNEL_MESSAGE_WITH_SOURCE, 
            'data': message._prepare().to_dict()
        }
        
        data = await self.http.request(
            'POST', 
            f'/interactions/{self.id}/{self.token}/callback', 
            data=content, 
            files=[fp.path for fp in message.attachments] if message.attachments else None,
            params={'with_response': with_response}
        )

        if with_response:
            return InteractionCallbackModel.from_dict(data)
        
    async def update(
        self,
        *,
        suppress_embeds: bool = None,
        **options: Unpack[EditMessageParams]
    ) -> None:
        """Edits the initial Interaction response.

        Args:
            options (EditMessageParams): fields to edit
            suppress_embeds (optional, bool): whether the response's embeds should be removed
        """
        options = serialize(options)
        self._apply_suppress_embeds(options, suppress_embeds)
        files = self._prepare_attachments(options)

        content = {
            "type": InteractionCallbackType.UPDATE_MESSAGE,
            "data": options,
        }

        await self.http.request(
            "POST",
            f"/interactions/{self.id}/{self.token}/callback",
            data=content,
            files=files
        )

    async def respond_modal(self, modal: ModalPart) -> None:
        """Create a modal in response to an interaction.
        Fires [`InteractionEvent`][scurrypy.events.interaction_events.InteractionEvent].

        Args:
            modal (ModalPart): modal data
        """
        content = {
            'type': InteractionCallbackType.MODAL,
            'data': modal.to_dict()
        }

        await self.http.request(
            'POST', 
            f'/interactions/{self.id}/{self.token}/callback', 
            data=content)

    async def respond_autocomplete(self, choices: list[CommandOptionChoicePart]) -> None:
        """Autocomplete a command in response to an interaction.
        Fires [`InteractionEvent`][scurrypy.events.interaction_events.InteractionEvent].

        Args:
            choices (list[CommandOptionChoicePart]): list of choices to autocomplete
        """
        content = {
            'type': InteractionCallbackType.APPLICATION_COMMAND_AUTOCOMPLETE_RESULT,
            'data': {
                'choices': [choice.to_dict() for choice in choices]
            }
        }

        await self.http.request(
            'POST',
            f'/interactions/{self.id}/{self.token}/callback',
            data=content
        )

    async def defer_respond(self, ephemeral: bool = None) -> None:
        """Defer creating a message in response to an interaction.
        Fires [`InteractionEvent`][scurrypy.events.interaction_events.InteractionEvent].

        Args:
            ephemeral (bool, optional): whether thinking + deferred interaction response is ephemeral
        """
        content = {
            'type': InteractionCallbackType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE,
            'data': {
                'flags': MessageFlags.EPHEMERAL if ephemeral else MessageFlags.NO_FLAGS
            }
        }

        await self.http.request(
            'POST',
            f'/interactions/{self.id}/{self.token}/callback',
            data=content
        )

    async def defer_update(self) -> None:
        """Defer updating a message in response to an interaction.
        Fires [`InteractionEvent`][scurrypy.events.interaction_events.InteractionEvent].
        """
        content = {
            'type': InteractionCallbackType.DEFERRED_UPDATE_MESSAGE,
        }

        await self.http.request(
            'POST',
            f'/interactions/{self.id}/{self.token}/callback',
            data=content
        )

    async def followup(
        self, 
        application_id: Snowflake, 
        message: str | MessagePart, 
        ephemeral: bool = None,
        suppress_embeds: bool = None
    ) -> None:
        """Create a new message to respond to a deferred interaction.
        Fires [`MessageCreateEvent`][scurrypy.events.message_events.MessageCreateEvent].

        !!! important
            Apps are limited to 5 followup messages PER interaction.

        Args:
            application_id (Snowflake): ID of the application
            message (str | MessagePart): content as a string or MessagePart
            ephemeral (optional, bool): whether the followup should be ephemeral
            suppress_embeds (optional, bool): whether the followup's embeds should be removed
        """
        if isinstance(message, str):
            message = MessagePart(content=message)

        message.flags = MessageFlags.NO_FLAGS

        if ephemeral:
            message.flags |= MessageFlags.EPHEMERAL

        if suppress_embeds:
            message.flags |= MessageFlags.SUPPRESS_EMBEDS

        content = message._prepare().to_dict()

        await self.http.request(
            'POST',
            f'/webhooks/{application_id}/{self.token}',
            data=content
        )

    async def edit_original(
        self,
        application_id: Snowflake,
        *,
        suppress_embeds: bool = None,
        **options: Unpack[EditMessageParams]
    ) -> None:
        """Edits the initial Interaction response.

        Args:
            application_id (Snowflake): bot's user ID
            options (EditMessageParams): fields to edit
            suppress_embeds (optional, bool): whether the response's embeds should be removed
        """
        options = serialize(options)
        self._apply_suppress_embeds(options, suppress_embeds)
        files = self._prepare_attachments(options)

        await self.http.request(
            "PATCH",
            f"/webhooks/{application_id}/{self.token}/messages/@original",
            data=options,
            files=files
        )



# scurrypy/api/commands

from .application_command import (
    ApplicationCommandOptionChoiceModel, 
    ApplicationCommandOptionModel,
    ApplicationCommandModel
)
from .context import (
    MessageCommandPart, 
    UserCommandPart
)
from .slash import (
    CommandOptionChoicePart, 
    CommandOptionPart, 
    SlashCommandPart
)

__all__ = [
    "ApplicationCommandOptionChoiceModel", 
    "ApplicationCommandOptionModel", 
    "ApplicationCommandModel",

    "MessageCommandPart", 
    "UserCommandPart",

    "CommandOptionChoicePart", 
    "CommandOptionPart", 
    "SlashCommandPart"
]
